# 分析设计文档

## quick start

- 构建compiler
    
    ```bash
    mkdir build
    cd build
    cmake ..
    make -j4
    ```
    
- 生成ast
    
    ```bash
    （仍然在build文件夹下，生成某个sy文件比如test.sy的ast）
    ./compiler --ast test.sy 
    ```
    
- 生成ir
    
    ```bash
    （仍然在build文件夹下，生成某个sy文件比如test.sy的ir）
    ./compiler --ir test.sy 
    ```
    
- 生成.s
    
    ```bash
    （仍然在build文件夹下，生成某个sy文件比如test.sy的.s，-O1选项会进行更多优化）
    ./compiler -S test.sy -o test.s
    ./compiler -S test.sy -o test.s -O1
    ```
    

## 前端

- 前端采用ANTLR框架构建词法分析器和语法分析器。输入的文件经过语法分析之后变成语法分析树，之后经过typer做类型检查之后进一步转换为中端IR表示。

## 中端

### 1. IR设计

- 数据结构
    - namespace ir
        - class DataMeta：保存变量的相关信息，包括数据类型、是否为全局变量、变量名、所对应的虚拟寄存器、初始值、是否能放入bss段等
        - class Instruction：所有指令的父类，保存一条指令所在基本块、指令类型、dsts、srcs
            - class Assign：赋值
            - class LoadImm4：加载立即数
            - class Unary：一元运算
            - class Binary：二元运算
            - class BinaryImm：有常量嵌入的二元运算
            - class Call：函数调用
            - class Branch：无条件跳转
            - class CondBranch：条件跳转
            - class Return：返回
            - class Alloca：分配空间
            - class Store：将数据存储到相应地址
            - class Load：读取相应地址的数据
            - class LoadAddr：读取全局变量所在地址
            - class ElementPtr：获取数组元素对应的指针
            - class Phi：phi语句，根据不同的前驱赋值不同的值
            - class Cast：用于整型与浮点数之间的类型转换
        - class BasicBlock：基本块，保存当前基本块的序号、所在function、所含指令，以及live_in、live_out、live_use等辅助信息
        - class Function：函数，保存返回值类型、所在module、参数列表、参数对应的虚拟寄存器、所含基本块、所含指令、已用的虚拟寄存器数量、已用的基本块数量
            - class LibFunction: public Function：库函数，是Function的子类，用于对Sysy2022里的库函数进行标记
        - class Module：程序，保存所含函数、全局变量等信息
    - Temp：虚拟寄存器，保存自身的序号以及所存数据的类型
    - Type：数据类型，基本类型有Int、Float、Void，如果是数组，则会记录每个维度的大小，如果是地址，则第一维是0
- 效果：将前端的ast翻译成→满足SSA形式，但有许多`alloca/load/store`指令的ir

### 2. 中端优化

- 数据/控制流分析
    - 支配关系分析：分析各个基本块之间的支配关系并建立支配树，得到支配边界等信息。
    - Use-Def分析：分析每个寄存器被定义和被使用的位置，从而有利于后续指令范围分析。
    - 循环分析：识别循环出现的位置，同时统计循环嵌套的次数，为循环相关优化提供关键信息。
- mem2reg
    - 优化思路参考《Static Single Assignment Book》和文章[https://buaa-se-compiling.github.io/miniSysY-tutorial/challenge/mem2reg/help.html](https://buaa-se-compiling.github.io/miniSysY-tutorial/challenge/mem2reg/help.html)，效果是识别出局部变量的 `alloca` 指令，将对应的局部变量改为虚拟寄存器中的 SSA 形式的变量，将该变量的 `store/load` 修改为虚拟寄存器之间的 def-use/use-def 关系，并在适当的地方加入 `phi` 指令和进行变量的重命名
- ArraySSA
    - 将地址访问相关指令进行数据流分析，让它们转化为访问寄存器的形式，从而可以进行消除冗余的地址访问以及优化访问位置，在优化结束后再重新转化为访问地址的指令。
- 常量传播
    - 效果是将常量沿数据流转播，将一些对常量的运算直接在编译阶段进行
    - 原理是找到LoadImm4指令就将对应常量记录到constant_map中，之后在Assign、Unary、Binary、CondBranch指令中判断，如果能进行常量的运算，就把结果也记录到constant_map中。整个过程直到没有新的常量记录到constant_map为止
- 全局值标号与全局代码移动
    - 优化参考[https://dl.acm.org/doi/pdf/10.1145/207110.207154](https://dl.acm.org/doi/pdf/10.1145/207110.207154)，同时实现全局值标号和全局代码移动作为一个组合优化。
    - 全局值标号根据一定计算规则对每个表达式分配一个值，并记录该值对应的寄存器，如果后续出现了相同的值，则直接使用赋值语句替换重复的指令，同时将后续使用到目的寄存器的位置全部替换为源寄存器。结合无用指令消除可以消除冗余计算。
    - 全局代码移动是根据寄存器间的依赖关系确定指令可以存在的安全范围，然后在这个范围中将指令尽可能移动到被调用次数少的基本块，这里一般尽可能选择循环嵌套次数更少的基本块。使用全局代码移动可以达到减少指令运行次数的目的。
- 常量嵌入
    - 效果是将操作数中含有一个常量的二元运算Binary，转变成有常量嵌入的二元运算BinaryImm，减少后端指令数
- 复写传播
    - 效果是消除掉ir中的所有冗余Assign指令
    - 原理是
        - 识别到一个Assign指令，如果它的src不在copy_map中，就将src加入copy_map，如果src已经在copy_map中，就令`copy_map[dst->get_index()] = copy_map[src->get_index()]`，达到传播的效果
        - 识别完之后，再遍历所有指令，把src在copy_map中的指令里的src全都改成`copy_map[src->get_index()]`
- 函数内联
    - 效果是将非递归函数内联到main函数中
    - 原理是
        - 先判断一个函数能否被内联，判断标准为：不是main函数、指令数小于10000、不调用自身
        - 再遍历main函数，找到一个Call指令且调用了可内联函数时，将Call指令之后的指令放入新的基本块，将原来的Call删除，并把要内联的函数的第一个基本块的内容插入到原来的Call指令位置。内联进来的基本块和虚拟寄存器的序号要在自身的基础上加上main函数的bb_used和temp_used
        - 整个过程直到无法再内联新函数为止
- 控制流化简
    - 优化思路参考《Engineering A Compiler》10.2.2节Eliminating Useless Control Flow的流程，目标是最简化控制流。
    - 每一个pass包含：（1）消除多余的跳转（2）消除多余的基本块（3）基本块融合（4）条件跳转提升；控制流化简会不断循环pass直到某个pass中没有进行任何控制流改变。
- 基本块排序
    - 基本原理是根据静态profile来估计块与块之间的branch probability和branch frequency，之后利用Pettis-Hansen Heuristic来进行块排序
    - 在我们的实现中，估计branch probability的时候我们根据块到块的不同情况（如循环中，跳出循环等）来使用事先确定的结果（常量）；例如我们认为循环中回到循环开始的概率比跳出循环的概率大，因此会给前者赋值为一个比较大的常量概率。思路和赋值常量参考了文章[**Static branch frequency and program profile analysis**](https://minds.wisconsin.edu/bitstream/handle/1793/59924/TR1248.pdf;jsessionid=15FD1B9025A6BCFDC0FE12578A685FEA?sequence=1)
- 循环展开
    - 识别出循环的上界、下界均为常量时，当展开后的指令条数在一定阈值内时，对其进行完全展开。
- 代数化简
    - 针对整数类型的加法和减法进行了优化，统计每个寄存器中各个值所累加的次数，当满足化简条件时将对应的运算指令转化为化简之后的指令，接着通过无用指令消除消除之前出现的冗余指令。
    - 当某些寄存器中存储的值参与累加并在后续的运算中被消减掉，该寄存器的计数器被标记为0，此时可以通过一条赋值语句将关于该寄存器的冗余计算消除掉。
    - 当一条寄存器中存储的值累加次数超过100次，可以通过一条乘法指令代替这100次加法指令，从而实现指令化简。
- 尾递归消除
    - 识别出尾递归，并将最后的Call指令改成跳转指令跳转到_B0，同时在_B0前插入新基本块，并在_B0的开头插入Phi语句，判断数据是来自参数函数来自尾递归
- 全局变量转局部
    - 对于只在main函数中的全局变量，将对其的操作改成对局部变量的操作，之后可以利用mem2reg消除`alloca/load/store`指令
    - 对于永远不会改变的全局变量，直接将其当做立即数
- 无用代码消除
    - 无用函数消除：找出main函数及其调用的函数中未使用到的函数，并将其删除
    - 无用指令消除：找出对`return/condbranch/call/store`指令没有影响的指令并删除
    - 无用循环消除：如果循环内的操作对循环外没有影响，则将其删除
- 强度削弱
    - 优化思路参考《Engineering A Compiler》10.7.2节 Strength Reduction的流程，目标是将循环中的乘法削弱成加法，并把无用的归纳变量删除

## 后端

### 1. 后端框架说明

后端主要完成 ir 到 riscv 机器码的生成过程，主要分为以下几个步骤：

- 指令选择：逐条翻译基本块内的指令，将其转换为相应的 riscv 机器指令。
- 指令调度：模拟多周期双发射流水线，按照一定启发规则对基本块内的指令进行重排。
- 生成 prologue 和 epilogue 块：计算栈帧大小，完成 fp, sp 在进入/离开函数时的设置，并在入口块种完成函数参数的加载。
- 寄存器分配：利用图染色寄存器分配算法，根据寄存器的生命周期和重要程度，启发式地分配寄存器。

### 2. 后端优化

- 指令翻译
    - Remove WAW/WAR hazards
        - 使翻译得到指令的目的寄存器尽可能不一样，有利于缩短寄存器生命周期，减轻寄存器分配压力，方便指令调度。
    - 乘常数/除常数优化
        - 根据开发板说明，一条 DIV/REM 指令的开销可能高达 68 个时钟周期。参考 [Division by invariant integers using multiplication | ACM SIGPLAN Notices](https://dl.acm.org/doi/10.1145/773473.178249) 论文中的算法，将整数除法转化为乘法和位移指令的组合。
    - x0 的嵌入：在指令某一源操作数为常数 0 的情况下，利用 x0 作为源操作数
    - sw → sd：识别到两次连续 sw 到某一连续地址时，将其转化为 sd，减少访存次数
    - beqz 转化：优化 ir 中对分支跳转的翻译，减少指令条数
    - 冗余指令消除：经过上述指令翻译，可能出现冗余指令，将其消除
- 指令调度
    - Software Pipelining
        - 模拟多周期双发射流水线 CPU 对指令的硬件执行过程，以结束执行所有指令花费时钟周期最小为目标，对基本块内的指令进行调度重排。
- 图染色寄存器分配
    - 参考 [Coloring-based coalescing for graph coloring register allocation | Proceedings of the 8th annual IEEE/ACM international symposium on Code generation and optimization](https://dl.acm.org/doi/10.1145/1772954.1772978) 论文中的算法，按照如下步骤进行图染色寄存器分配：
        1. build：经过数据流分析，同时出现的虚拟寄存器之间连边，move 的 src 和 dst 之间连边
        2. simplify + coalesce：
            1. simplify：去除的点为非 move related 且度小于 K 的点
            2. coalesce：move 的 src 和 dst 节点相邻的 significant degree（度数大于 K）的节点数小于 K 时，可以进行合并，同时删除 move 指令；交替 simplify 和 coalesce 直到图中只剩下 move related 节点和 significant degree 节点
        3. freeze：放弃一个 move related 节点，将其标记为非 move related，回到步骤 2
        4. select：按照节点存在栈上的顺序分配寄存器
    - 在考虑选择虚拟寄存器 spill 到栈上时，考虑其是否为加载代价较低的 li 或 la 或与 sp 相关的 addi 指令，从而选择 spill 代价最小的寄存器
- 控制流优化
    - 基本思路与中端的控制流化简一致，但是由于是在翻译指令后进行所以没有分支提升优化

## 测试框架

- 配置了gitlab的cicd，采用x86服务器充当前端-riscv星光开发板充当后端的前后端设计，具体运行环境和过程如下：
    - gitlab runner运行在服务器上，测例和测试程序运行在开发板上
    - cicd开始后，服务器向开发板发送commit数据，开发板启动测试，从仓库拉取对应commit的版本并运行所有测例，测试数据返回给服务器，进而显示在gitlab上。